#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# License: GNU General Public License v2
#
# Author: thl-cmk[at]outlook[dot]com
# URL   : https://thl-cmk.hopto.org
# Date  : 2024-07-05
# File  : vsphere_topo/agent_based/packages.py

# 2024-07-10: added time taken as metric
# 2024-07-11: added clearer error message for missing section_esx_vsphere_clusters section
#             added automatic workaround for backend is only picking up topologies from default folder
#             -> obsoletes 'Add dummy topologies'
#             added vCenter twice (once as vCenter once as vM) with different object IDs
#             -> obsoletes 'Don\'t add vCenter as VM'
# 2024-07-12: added support for Queried host is a host system

from collections.abc import Mapping, Sequence
from pathlib import Path
from time import time_ns
from typing import Any

from cmk.agent_based.v2 import (
    CheckPlugin,
    CheckResult,
    DiscoveryResult,
    Result,
    Service,
    State,
    check_levels,
    render,
)
from cmk.base.check_api import host_name
from cmk_addons.plugins.vsphere_topo.lib.utils import (
    EMBLEM_CLUSTER,
    EMBLEM_DATA_CENTER,
    EMBLEM_DATA_STORE,
    ICON_VCENTER,
    LiveStatusConnection,
    OMD_ROOT,
    PARAM_CLUSTER,
    PARAM_DATA_CENTER,
    PARAM_DATA_STORE,
    PARAM_DATA_STORE_AS_SERVICE,
    PARAM_HOST_SYSTEMS,
    PARAM_MAKE_DEFAULT,
    PARAM_VCENTER,
    PARAM_VM_NAMES,
    RULE_SET_NAME_VSPHERE_TOPO,
    adjust_name,
    get_emblem,
    get_topologies,
    save_data_to_file,
)

Section = Mapping[str, object]


def discover_vsphere_topo(
        section_esx_vsphere_clusters: Section | None,
        section_esx_vsphere_virtual_machines: Sequence[Section] | None,
        section_esx_vsphere_datastores: Section | None,
) -> DiscoveryResult:
    yield Service()


def check_vsphere_topo(
        params: Mapping[str, Any],
        section_esx_vsphere_clusters: Section | None,
        section_esx_vsphere_virtual_machines: Sequence[Section] | None,
        section_esx_vsphere_datastores: Section | None,
) -> CheckResult:
    def add_host(
            host: str,
            emblem: str | None = None,
            icon: str | None = None,
            link2core: bool = True,
            obj_id_prefix: str = '',
    ):
        if objects.get(f'{obj_id_prefix}{host}') is not None:
            return

        metadata = {}
        if emblem or icon:
            metadata = {'images': {}}

        if emblem is not None:
            metadata['images'].update({'emblem': emblem})  # node image

        if icon is not None:
            metadata['images'].update({'icon': icon})  # node icon

        objects[f'{obj_id_prefix}{host}'] = {
            "name": host,
            "link": {'core': host} if link2core else {},
            'metadata': metadata,
        }

    def add_service(host: str, service: str, emblem: str | None = None, link2core: bool = True):
        obj_id = f'{service}@{host}'
        if objects.get(obj_id) is not None:
            return

        metadata = {}
        if emblem is not None:
            metadata = {
                'images': {
                    'emblem': emblem,  # node image
                },
            }

        objects[obj_id] = {
            "name": service,
            "link": {'core': [host, service]} if link2core else {},
            'metadata': metadata,
        }

    def add_connection(left: str, right: str):
        connection = [left, right]
        connection.sort()
        connections.append([connection, {'line_config': {'css_styles': {'stroke-dasharray': 'unset'}}}])

    start_time = time_ns()
    objects = {}
    connections = []

    if section_esx_vsphere_clusters is not None:
        __clusters = {
            'HA': {
                'datacenter': 'datacenter-21',
                'hostsystems': 'esxi01, esxi02',
                'vms': 'vmname01, vmname02, vmname03'
            }
        }

        raw_vsphere_host: str = host_name().strip()
        add_host(
            host=raw_vsphere_host,
            obj_id_prefix='vc',
            icon=get_emblem(ICON_VCENTER, params.get(PARAM_VCENTER)),
        )
        vsphere_host = f'vc{raw_vsphere_host}'

        for cluster in section_esx_vsphere_clusters:
            data_center: str = section_esx_vsphere_clusters[cluster]['datacenter'].strip()
            if objects.get(data_center) is None:
                add_host(
                    host=data_center,
                    emblem=get_emblem(EMBLEM_DATA_CENTER, params.get(PARAM_DATA_CENTER)),
                    link2core=False,
                )
                add_connection(data_center, vsphere_host)
            add_host(
                host=cluster,
                emblem=get_emblem(EMBLEM_CLUSTER, params.get(PARAM_CLUSTER)),
                link2core=False
            )
            add_connection(cluster, data_center)

            host_systems: Sequence[str] = section_esx_vsphere_clusters[cluster].get('hostsystems', '').split(',')
            for host_system in host_systems:
                esx_name: str = adjust_name(host_system, params.get(PARAM_HOST_SYSTEMS, {}))
                add_host(host=esx_name)
                add_connection(cluster, esx_name)
        yield Result(
            state=State.OK,
            notice='VMware ESX via vSphere agent type of query: Queried host is the vCenter'
        )
    else:  # assuming the query type is ESXi host
        raw_vsphere_host: str = host_name().strip()
        add_host(host=raw_vsphere_host)
        vsphere_host = raw_vsphere_host
        yield Result(
            state=State.OK,
            notice='VMware ESX via vSphere agent type of query: Queried host is a host system'
        )

    if section_esx_vsphere_virtual_machines is not None:
        __VMs = [
            {
                'vm_name': 'vmname01',
                'hostsystem': 'esxi01',
                'powerstate': 'poweredOff',
                'guest_os': 'Microsoft Windows Server 2008 R2 (64-bit)',
                'compatibility': 'vmx-10',
                'uuid': '420354bd-bee0-88d5-053c-d4f424106b0f'
            },
            {
                'vm_name': 'vmname02',
                'hostsystem': 'esxi02',
                'powerstate': 'poweredOn',
                'guest_os': 'Microsoft Windows Server 2019 (64-bit)',
                'compatibility': 'vmx-19',
                'uuid': '422eba9f-0327-2da0-b546-f5a5e5dfdb68'
            },
            {
                'vm_name': 'vmname03',
                'hostsystem': 'esxi01',
                'powerstate': 'poweredOn',
                'guest_os': 'Other 3.x or later Linux (64-bit)',
                'compatibility': 'vmx-11',
                'uuid': '422e3863-b7de-dc55-74f8-6dc9c159b36d'
            }
        ]
        for vm in section_esx_vsphere_virtual_machines:
            vm_name: str = adjust_name(vm['vm_name'], params.get(PARAM_VM_NAMES, {}))
            host_system: str = adjust_name(vm['hostsystem'], params.get(PARAM_HOST_SYSTEMS, {}))
            add_host(host=vm_name)
            add_connection(vm_name, host_system)

    service_prefix: str = ''
    ds_services = None
    ls_connection = LiveStatusConnection()

    if section_esx_vsphere_datastores is not None:
        data_stores_host: str = 'data_stores'  # Anchor for data stores
        add_host(
            host=data_stores_host,
            link2core=False,
            emblem=get_emblem(EMBLEM_DATA_STORE, params.get(PARAM_DATA_STORE)),
        )
        for ds_name in section_esx_vsphere_datastores:
            add_host(
                host=ds_name,
                link2core=False,
                emblem=get_emblem(EMBLEM_DATA_STORE, params.get(PARAM_DATA_STORE)),
            )
            add_connection(ds_name, data_stores_host)

        if params.get(PARAM_DATA_STORE_AS_SERVICE) is True:
            query: str = (
                'GET services\n'
                'Columns: service_description\n'
                'Filter: service_check_command ~ ^check_mk-esx_vsphere_datastores(!.*)?\n'
                f'Filter: host_name = {vsphere_host}\n'
                'OutputFormat: python3\n'
            )
            if (ds_services := ls_connection.query(query=query)) is not None:
                for ds_service in ds_services:
                    service_name: str = ds_service[0]
                    service_id: str = f'{service_name}@{vsphere_host}'
                    ds_name: str = service_name.split('fs_')[-1].split('Filesystem ')[-1]
                    add_service(
                        host=vsphere_host,
                        service=service_name,
                    )
                    add_connection(service_id, ds_name)
                    service_prefix = service_name.split(ds_name)[0]

    query: str = (
        'GET services\n'
        'Columns: host_name long_plugin_output\n'
        'Filter: description = ESX Datastores\n'
        'OutputFormat: python3\n'
    )
    if (vm_to_data_stores := ls_connection.query(query=query)) is not None:
        for vm, data_stores in vm_to_data_stores:
            vm_name: str = adjust_name(vm, params.get(PARAM_VM_NAMES, {}))
            if vm_name in objects:
                data_stores = data_stores.split(',')
                for data_store in data_stores:
                    ds_name = data_store.strip().split(' ')[2]
                    if ds_services is not None:
                        service_name = f'{service_prefix}{ds_name}'
                        service_id = f'{service_name}@{vsphere_host}'
                        add_connection(service_id, vm_name)
                    else:
                        add_connection(ds_name, vm_name)

    connections.sort()
    data_set_vsphere = {
        'version': 1,
        'name': 'vSphere',
        'objects': dict(sorted(objects.items())),
        'connections': connections,
    }

    data_sets = []

    # workaround for backend is only picking up topologies from default folder
    path: str = f'{OMD_ROOT}/var/check_mk/topology/data/default'
    if Path(path).exists() and not Path(f'{path}/data_vsphere.json').exists():  # don't overwrite existing topology
        dummy_topology = {'version': 1, 'name': 'vSphere', 'objects': {}, 'connections': []}
        save_data_to_file(
            data=dummy_topology,
            file='data_vsphere.json',
            make_default=params.get(PARAM_MAKE_DEFAULT, False),
            path=path,
        )

    for topology in get_topologies():
        data_sets.append({'version': 1, 'name': topology, 'objects': {}, 'connections': []})
    # end workaround

    data_sets.append(data_set_vsphere)

    path: str = f'{OMD_ROOT}/var/check_mk/topology/data/{raw_vsphere_host}'

    for data_set in data_sets:
        file: str = f'data_{data_set["name"].lower()}.json'
        save_data_to_file(
            data=data_set,
            file=file,
            make_default=params.get(PARAM_MAKE_DEFAULT, False),
            path=path,
        )

    yield Result(state=State.OK, summary=f'Objects: {len(objects)}')
    yield Result(state=State.OK, summary=f'Connections: {len(connections)}')
    yield Result(state=State.OK, notice=f'Written to: {path}/{file}')
    yield from check_levels(
        value=(time_ns() - start_time) / 1e9,
        label='Time taken',
        metric_name='vsphere_topo_time_taken',
        boundaries=(0, None),
        render_func=render.timespan,
    )


check_plugin_vsphere_topo = CheckPlugin(
    name="vsphere_topo",
    service_name="vSphere Topology",
    sections=[
        'esx_vsphere_clusters',
        'esx_vsphere_virtual_machines',
        'esx_vsphere_datastores',
    ],
    check_function=check_vsphere_topo,
    discovery_function=discover_vsphere_topo,
    check_ruleset_name=RULE_SET_NAME_VSPHERE_TOPO,
    check_default_parameters={},
)
